"""
Methods for behavioral-epoch detection and analysis
"""

import numpy as np
import pandas as pd
from scipy.signal import find_peaks

from ..aux import nam
from .. import reg, aux, util

__all__ = [
    'process_epochs',
    'stride_interp',
    'detect_strides',
    # 'weathervanesNheadcasts',
    # 'comp_epochs_byID',
    # 'bout_distribution',
    # 'bout_detection',
    # 'compute_interference_data',
    'compute_interference',
    # 'turn_mode_annotation',
    # 'turn_annotation',
    # 'crawl_annotation',
]











def process_epochs(a, epochs, dt, return_idx=False):
    if epochs.shape[0] == 0:
        stops = np.array([])
        durs = np.array([])
        slices = []
        amps = np.array([])
        idx = np.array([])
        maxs = np.array([])
        if return_idx:
            return stops, durs, slices, amps, idx, maxs
        else:
            return durs, amps, maxs

    else:
        if epochs.shape == (2,):
            epochs = np.array([epochs, ])
        durs = (np.diff(epochs).flatten()) * dt
        slices = [np.arange(r0, r1, 1) for r0, r1 in epochs]
        amps = np.array([np.trapz(a[p][~np.isnan(a[p])], dx=dt) for p in slices])
        maxs = np.array([np.max(a[p]) for p in slices])
        if return_idx:
            stops = epochs[:, 1]
            idx = np.concatenate(slices) if len(slices) > 1 else slices[0]
            return stops, durs, slices, amps, idx, maxs
        else:
            return durs, amps, maxs




def detect_strides(a, dt, vel_thr=0.3, stretch=(0.75, 2.0), fr=None, return_extrema=True, return_runs=True):
    """
    Annotates strides-runs and pauses in timeseries.

    Extended description of function.

    Parameters
    ----------
    a : array
        1D np.array : forward velocity timeseries
    dt : float
        Timestep of the timeseries
    vel_thr : float
        Maximum velocity threshold
    stretch : Tuple[float,float]
        The min-max stretch of a stride relative to the default derived from the dominnt frequency
    fr : float, optional
        The dominant crawling frequency.
    return_extrema : boolean
        Whether to additionally return the stride extrema
    return_runs : boolean
        Whether to additionally return the runs (stridechains)

    Returns
    -------
    strides : list
        A list of pairs of the start-end indices of the strides.
    i_min : array
        Indices of the local minima.
    i_max : array
        Indices of the local maxima
    runs : list
         A list of pairs of the start-end indices of the runs/stridechains.
    run_counts : list
         Stride-counts of the runs/stridechains.

    """
    if fr is None:
        fr = aux.fft_max(a, dt, fr_range=(1, 2.5))
    tmin = stretch[0] // (fr * dt)
    tmax = stretch[1] // (fr * dt)
    i_min = find_peaks(-a, height=-3 * vel_thr, distance=tmin)[0]
    i_max = find_peaks(a, height=vel_thr, distance=tmin)[0]
    strides = []
    for m in i_max:
        try:
            s0, s1 = [i_min[i_min < m][-1], i_min[i_min > m][0]]
            if ((s1 - s0) <= tmax) and ([s0, s1] not in strides):
                strides.append([s0, s1])
        except:
            pass
    strides = np.array(strides)
    if not return_runs:
        if return_extrema:
            return i_min, i_max, strides
        else:
            return strides

    runs, run_counts = [], []
    s00, s11 = None, None

    count = 0
    for ii, (s0, s1) in enumerate(strides.tolist()):
        if ii == 0:
            s00, s11 = s0, s1
            count = 1
            continue
        if s11 == s0:
            s11 = s1
            count += 1
        else:
            runs.append([s00, s11])
            run_counts.append(count)
            count = 1
            s00, s11 = s0, s1
        if ii == len(strides) - 1:
            runs.append([s00, s11])
            run_counts.append(count)
            break
    runs = np.array(runs)
    if return_extrema:
        return i_min, i_max, strides, runs, run_counts
    else:
        return strides, runs, run_counts





def stride_interp(a, strides, Nbins=64):
    x = np.linspace(0, 2 * np.pi, Nbins)
    aa = np.zeros([strides.shape[0], Nbins])
    for ii, (s0, s1) in enumerate(strides):
        aa[ii, :] = np.interp(x, np.linspace(0, 2 * np.pi, s1 - s0), a[s0:s1])
    return aa


def mean_stride_curve(a, strides, da, Nbins=64):
    aa = stride_interp(a, strides, Nbins)
    aa_minus = aa[da < 0]
    aa_plus = aa[da > 0]
    aa_norm = np.vstack([aa_plus, -aa_minus])
    dic = aux.AttrDict({
        'abs': np.nanquantile(np.abs(aa), q=0.5, axis=0).tolist(),
        'plus': np.nanquantile(aa_plus, q=0.5, axis=0).tolist(),
        'minus': np.nanquantile(aa_minus, q=0.5, axis=0).tolist(),
        'norm': np.nanquantile(aa_norm, q=0.5, axis=0).tolist(),
    })

    return dic


def cycle_curve_dict(s, dt, shs=['sv', 'fov', 'rov', 'foa', 'b']):
    strides = detect_strides(s[reg.getPar('sv')], dt, return_extrema=False, return_runs=False)
    da = np.array([np.trapz(s[reg.getPar('fov')][s0:s1].dropna()) for ii, (s0, s1) in enumerate(strides)])
    dic = {sh: mean_stride_curve(s[reg.getPar(sh)], strides, da) for sh in shs}
    return aux.AttrDict(dic)


def cycle_curve_dict_multi(s, dt, shs=['sv', 'fov', 'rov', 'foa', 'b']):
    # ssv=s[reg.getPar('sv')]
    # sfov=s[reg.getPar('fov')]
    # ps=reg.getPar(shs)
    # sshs=aux.AttrDict({sh:s[reg.getPar(sh)] for sh in shs})

    ids = s.index.unique('AgentID').values
    dic = {}
    for id in ids:
        ss = s.xs(id, level="AgentID")
        dic[id] = cycle_curve_dict(ss, dt=dt, shs=shs)
    return aux.AttrDict(dic)


def compute_interference(s, e, c, d=None, Nbins=64, chunk_dicts=None):
    p_sv, p_fov, p_rov, p_foa, p_b, pau_fov_mu = reg.getPar(['sv', 'fov', 'rov', 'foa', 'b', 'pau_fov_mu'])

    x = np.linspace(0, 2 * np.pi, Nbins)

    sss = {id: s.xs(id, level="AgentID") for id in c.agent_ids}

    if chunk_dicts is None:
        stride_dic = {}

        stride_dic_dfo = {}
        for jj, id in enumerate(c.agent_ids):
            ss = sss[id]
            stride_dic[id] = detect_strides(ss[p_sv].values, c.dt, return_runs=False, return_extrema=False)
            a_fov = ss[p_fov].values
            stride_dic_dfo[id] = np.array([np.trapz(a_fov[s0:s1]) for ii, (s0, s1) in enumerate(stride_dic[id])])
    else:
        stride_dic = {id: chunk_dicts[id]['stride'] for id in c.agent_ids}
        stride_dic_dfo = {id: chunk_dicts[id]['stride_Dor'] for id in c.agent_ids}

    # pooled_curves = {}
    cycle_curves = {}
    mean_curves_abs = {}
    for sh in ['sv', 'fov', 'rov', 'foa', 'b']:
        par = reg.getPar(sh)
        curves_abs = np.zeros([c.N, Nbins]) * np.nan
        curves_plus = np.zeros([c.N, Nbins]) * np.nan
        curves_minus = np.zeros([c.N, Nbins]) * np.nan
        curves_norm = np.zeros([c.N, Nbins]) * np.nan
        for jj, id in enumerate(c.agent_ids):
            ss = sss[id]
            aa = stride_interp(ss[par].values, stride_dic[id], Nbins=64)
            aa_plus = aa[stride_dic_dfo[id] > 0]
            aa_minus = aa[stride_dic_dfo[id] < 0]
            aa_norm = np.vstack([aa_plus, -aa_minus])
            curves_abs[jj, :] = np.nanquantile(np.abs(aa), q=0.5, axis=0)
            curves_plus[jj, :] = np.nanquantile(aa_plus, q=0.5, axis=0)
            curves_minus[jj, :] = np.nanquantile(aa_minus, q=0.5, axis=0)
            curves_norm[jj, :] = np.nanquantile(aa_norm, q=0.5, axis=0)
        mean_curves_abs[sh] = curves_abs
        cycle_curves[sh] = aux.AttrDict({
            'abs': curves_abs,
            'plus': curves_plus,
            'minus': curves_minus,
            'norm': curves_norm,
        })
        # pooled_curves[sh] = aux.AttrDict({
        #     'abs': np.nanquantile(curves_abs, q=0.5, axis=0).tolist(),
        #     'plus': np.nanquantile(curves_plus, q=0.5, axis=0).tolist(),
        #     'minus': np.nanquantile(curves_minus, q=0.5, axis=0).tolist(),
        #     'norm': np.nanquantile(curves_norm, q=0.5, axis=0).tolist(),
        # })

    att0s, att1s = np.min(mean_curves_abs['fov'], axis=1), np.max(mean_curves_abs['fov'], axis=1)

    e[aux.nam.phi(aux.nam.max('attenuation'))] = x[np.argmax(mean_curves_abs['fov'], axis=1)]
    e[aux.nam.phi(aux.nam.max(p_sv))] = x[np.argmax(mean_curves_abs['sv'], axis=1)]
    e[reg.getPar('str_sv_max')] = np.max(mean_curves_abs['sv'], axis=1)
    try:
        e['attenuation'] = att0s / e[pau_fov_mu]
        e[aux.nam.max('attenuation')] = (att1s - att0s) / e[pau_fov_mu]
    except:
        pass
    return cycle_curves


# @reg.funcs.annotation("turn")
# def turn_annotation(s, e, c):
#     eTur_ps = reg.getPar(['Ltur_N', 'Rtur_N', 'tur_N', 'tur_H'])
#     eTur_vs = np.zeros([c.N, len(eTur_ps)]) * np.nan
#     turn_ps = reg.getPar(['tur_fou', 'tur_t', 'Ltur_t', 'Rtur_t', 'tur_fov_max'])
#     turn_vs = np.zeros([c.Nticks, c.N, len(turn_ps)]) * np.nan
#     turn_dict = {}
#
#     for jj, id in enumerate(c.agent_ids):
#         a_fov = s[reg.getPar('fov')].xs(id, level="AgentID")
#         Lturns, Rturns = detect_turns(a_fov, c.dt)
#
#         Lturns1, Ldurs, Lturn_slices, Lamps, Lturn_idx, Lmaxs = process_epochs(a_fov.values, Lturns, c.dt,
#                                                                                return_idx=True)
#         Rturns1, Rdurs, Rturn_slices, Ramps, Rturn_idx, Rmaxs = process_epochs(a_fov.values, Rturns, c.dt,
#                                                                                return_idx=True)
#         Lturns_N, Rturns_N = Lturns.shape[0], Rturns.shape[0]
#         turns_N = Lturns_N + Rturns_N
#         tur_H = Lturns_N / turns_N if turns_N != 0 else 0
#
#         if Lturns_N > 0:
#             turn_vs[Lturns[:, 1], jj, 0] = Lamps
#             turn_vs[Lturns[:, 1], jj, 1] = Ldurs
#             turn_vs[Lturns[:, 1], jj, 2] = Ldurs
#             turn_vs[Lturns[:, 1], jj, 4] = Lmaxs
#         if Rturns_N > 0:
#             turn_vs[Rturns[:, 1], jj, 0] = Ramps
#             turn_vs[Rturns[:, 1], jj, 1] = Rdurs
#             turn_vs[Rturns[:, 1], jj, 3] = Rdurs
#             turn_vs[Rturns[:, 1], jj, 4] = Rmaxs
#         turn_dict[id] = {'Lturn': Lturns, 'Rturn': Rturns, 'turn_slice': Lturn_slices + Rturn_slices,
#                          'turn_amp': np.concatenate([Lamps, Ramps]),
#                          'Lturn_amp': Lamps, 'Rturn_amp': Ramps,
#                          'turn_dur': np.concatenate([Ldurs, Rdurs]), 'Lturn_dur': Ldurs, 'Rturn_dur': Rdurs,
#                          'turn_vel_max': np.concatenate([Lmaxs, Rmaxs])}
#         eTur_vs[jj, :] = [Lturns_N, Rturns_N, turns_N, tur_H]
#     s[turn_ps] = turn_vs.reshape(-1, len(turn_ps))
#     e[eTur_ps] = eTur_vs
#
#     return turn_dict
#
#
# @reg.funcs.annotation("crawl")
# def crawl_annotation(s, e, c, strides_enabled=True, vel_thr=0.3):
#     l, v, sv, dst, acc, fov, foa, b, bv, ba, fv = \
#         reg.getPar(['l', 'v', 'sv', 'd', 'a', 'fov', 'foa', 'b', 'bv', 'ba', 'fv'])
#
#     str_ps = reg.getPar(['str_d_mu', 'str_d_std', 'str_sv_mu', 'str_fov_mu', 'str_fov_std', 'str_N'])
#     lin_ps = reg.getPar(
#         ['run_v_mu', 'pau_v_mu', 'run_a_mu', 'pau_a_mu', 'run_fov_mu', 'run_fov_std', 'pau_fov_mu', 'pau_fov_std',
#          'run_foa_mu', 'pau_foa_mu', 'pau_b_mu', 'pau_b_std', 'pau_bv_mu', 'pau_bv_std', 'pau_ba_mu', 'pau_ba_std',
#          'cum_run_t', 'cum_pau_t', 'run_t_min', 'run_t_max', 'pau_t_min', 'pau_t_max'])
#
#     dt = c.dt
#     lin_vs = np.zeros([c.N, len(lin_ps)]) * np.nan
#     str_vs = np.zeros([c.N, len(str_ps)]) * np.nan
#
#     run_ps = reg.getPar(['pau_t', 'run_t', 'run_d', 'str_c_l', 'str_t', 'str_d', 'str_sd'])
#     run_vs = np.zeros([c.Nticks, c.N, len(run_ps)]) * np.nan
#
#     crawl_dict = {}
#
#     for jj, id in enumerate(c.agent_ids):
#         sv_minima, sv_maxima, strides, str_chain_ls, stride_Dor, stride_durs, stride_dsts, stride_sdsts, strides1 = [], [], [], [], [], [], [], [], []
#         a_v = s[v].xs(id, level="AgentID").values
#         a_fov = s[fov].xs(id, level="AgentID").values
#
#         if c.Npoints > 1:
#             a_sv = s[sv].xs(id, level="AgentID").values
#             if strides_enabled:
#                 sv_minima, sv_maxima, strides, runs, str_chain_ls = detect_strides(a_sv, dt, fr=e[fv].loc[id],
#                                                                                    vel_thr=vel_thr,
#                                                                                    return_extrema=True)
#                 strides1, stride_durs, stride_slices, stride_dsts, stride_idx, stride_maxs = process_epochs(a_v,
#                                                                                                             strides, dt,
#                                                                                                             return_idx=True)
#                 stride_sdsts = stride_dsts / e[l].loc[id]
#                 stride_Dor = np.array([np.trapz(a_fov[s0:s1 + 1]) for s0, s1 in strides])
#                 if stride_idx != []:
#                     str_fovs = np.abs(a_fov[stride_idx])
#                     str_vs[jj, :] = [np.nanmean(stride_dsts),
#                                      np.nanstd(stride_dsts),
#                                      np.nanmean(a_sv[stride_idx]),
#                                      np.nanmean(str_fovs),
#                                      np.nanstd(str_fovs),
#                                      np.nansum(str_chain_ls),
#                                      ]
#             else:
#                 runs = detect_runs(a_sv, dt, vel_thr=vel_thr)
#             pauses = detect_pauses(a_sv, dt, vel_thr=vel_thr, runs=runs)
#         else:
#
#             runs = detect_runs(a_v, dt, vel_thr=vel_thr)
#             pauses = detect_pauses(a_v, dt, runs=runs, vel_thr=vel_thr)
#
#         pauses1, pause_durs, pause_slices, pause_dsts, pause_idx, pause_maxs = process_epochs(a_v, pauses, dt,
#                                                                                               return_idx=True)
#         runs1, run_durs, run_slices, run_dsts, run_idx, run_maxs = process_epochs(a_v, runs, dt,
#                                                                                   return_idx=True)
#         if pauses1 != []:
#             run_vs[pauses1, jj, 0] = pause_durs
#         if strides1 != []:
#             run_vs[runs1, jj, 1] = run_durs
#             run_vs[runs1, jj, 2] = run_dsts
#             run_vs[runs1, jj, 3] = str_chain_ls
#         if strides1 != []:
#             run_vs[strides1, jj, 4] = stride_durs
#             run_vs[strides1, jj, 5] = stride_dsts
#             run_vs[strides1, jj, 6] = stride_sdsts
#
#         if b in s.columns and pause_idx != []:
#             pau_bs = s[b].xs(id, level="AgentID").abs().values[pause_idx]
#             pau_bvs = s[bv].xs(id, level="AgentID").abs().values[pause_idx]
#             pau_bas = s[ba].xs(id, level="AgentID").abs().values[pause_idx]
#             pau_b_temp = [np.mean(pau_bs), np.std(pau_bs), np.mean(pau_bvs), np.std(pau_bvs), np.mean(pau_bas),
#                           np.std(pau_bas)]
#         else:
#             pau_b_temp = [np.nan] * 6
#         a_foa = s[foa].xs(id, level="AgentID").abs().values
#         a_acc = s[acc].xs(id, level="AgentID").values
#
#         if run_idx != []:
#             run_fovs = np.abs(a_fov[run_idx])
#             run_foas = a_foa[run_idx]
#             run_vvs = a_v[run_idx]
#             run_aas = a_acc[run_idx]
#         else:
#             run_fovs, run_foas = [], []
#             run_vvs, run_aas = [], []
#         if pause_idx != []:
#             pau_fovs = np.abs(a_fov[pause_idx])
#             pau_foas = a_foa[pause_idx]
#             pau_vvs = a_v[pause_idx]
#             pau_aas = a_acc[pause_idx]
#         else:
#             pau_fovs, pau_foas = [], []
#             pau_vvs, pau_aas = [], []
#
#         lin_vs[jj, :] = [
#             np.mean(run_vvs),
#             np.mean(pau_vvs),
#             np.mean(run_aas),
#             np.mean(pau_aas),
#             np.mean(run_fovs), np.std(run_fovs),
#             np.mean(pau_fovs), np.std(pau_fovs),
#             np.mean(run_foas),
#             np.mean(pau_foas),
#             *pau_b_temp,
#             np.sum(run_durs),
#             np.sum(pause_durs),
#             np.nanmin(run_durs) if len(run_durs) > 0 else 1,
#             np.nanmax(run_durs) if len(run_durs) > 0 else 100,
#             np.nanmin(pause_durs) if len(pause_durs) > 0 else dt,
#             np.nanmax(pause_durs) if len(pause_durs) > 0 else 100,
#         ]
#         crawl_dict[id] = {'vel_minima': sv_minima, 'vel_maxima': sv_maxima, 'stride': strides, 'stride_Dor': stride_Dor,
#                           'exec': runs, 'pause': pauses,
#                           'run_idx': run_idx, 'pause_idx': pause_idx, 'stride_dur': stride_durs,
#                           'run_count': str_chain_ls, 'run_dur': run_durs, 'run_dst': run_dsts, 'pause_dur': pause_durs}
#     s[run_ps] = run_vs.reshape(-1, len(run_ps))
#     e[lin_ps] = lin_vs
#
#     str_d_mu, str_d_std, str_sd_mu, str_sd_std, run_tr, pau_tr, cum_run_t, cum_pau_t, cum_t = \
#         reg.getPar(
#             ['str_d_mu', 'str_d_std', 'str_sd_mu', 'str_sd_std', 'run_tr', 'pau_tr', 'cum_run_t', 'cum_pau_t', 'cum_t'])
#
#     e[run_tr] = e[cum_run_t] / e[cum_t]
#     e[pau_tr] = e[cum_pau_t] / e[cum_t]
#
#     if c.Npoints > 1 and strides_enabled:
#         e[str_ps] = str_vs
#         e[str_sd_mu] = e[str_d_mu] / e[l]
#         e[str_sd_std] = e[str_d_std] / e[l]
#     # print(s.columns)
#     return crawl_dict
